(in-package #:org-syntax)

(defclass lang ()
  ((block-lang :initarg :block-lang)))

(defvar empty-readtable
  (let ((*readtable*
          ;; There is no standard way to obtain the default readtable
          ;; so we just hope it's that.
          ;; Actually, here,
          ;; any readtable with #\Space being a whitespace character suffices.
          (copy-readtable)))
    (loop for c from 0 below char-code-limit
          do (set-syntax-from-char (code-char c) #\Space))
    *readtable*))

#+akater-types
(deftype header-args-alist () `(alist keyword t))
(deftype header-level () `(and fixnum (integer 0)))

#+akater-types
(declaim (type header-args-alist #:dummy)
         (type (alist header-level header-args-alist) #:dummy)
         (type (alist string (alist header-level header-args-alist)) #:dummy)
         (type (alist header-level t) #:dummy))

(deftype column-number () `(and (integer 0)
                                ;; We presume there are no long lines
                                fixnum))

;; A context is the state of a stream being read in org syntax
(defstruct context
  (now-reading nil :type (member nil properties block))
  (current-level 0 :type (integer 0))
  (header-args-stack nil
   ;; todo: rename to global-header-args-stack
   ;; All header args and depth at which they were set.
   ;; Always in descending order by car, e.g. ((6 ..) (4 ..) (3 ..) .. (0))
   #+akater-types :type
   #+akater-types (alist header-level header-args-alist))
  (langwise-header-args-stacks nil
   #+akater-types :type
   #+akater-types (alist string (alist header-level header-args-alist)))
  (todo-keywords-stack nil
   ;; All header args and depth at which they were set.
   ;; Always in descending order by car, e.g. ((6 ..) (4 ..) (3 ..) .. (0))
   #+akater-types :type
   #+akater-types (alist header-level t)
   ;; :type (alist header-level string)
   )
  ;; (block-headers-stack nil
  ;;  ;; blocks within blocks will rarely occur but they can occur
  ;;  ;; Still, I'm not sure we need this
  ;;  ;; this is only relevant for begin_src org
  ;;  ;; :type (list-of string)
  ;;  )
  (left-whitespace-p t :type boolean)
  (left-whitespace-column 0 :type column-number
   ;; non-nil iff the region between the latest newline (or the min position)
   ;; and current position does not contain non-whitespace characters
   ;; and if yes, it's the length of that region
   )
  (current-header-args nil
   ;; precomputed
   ;; :type header-args-alist
   )
  (current-langwise-header-args nil
   ;; precomputed
   ;; :type (alist lang header-args-alist)
   )
  (current-todo-keyword nil
   ;; precomputed
   ;; :type string
   )
  (current-todo-keywords-labelling-ignored-content '("TODO" "BUG" "HOLD"
                                                     "DEPRECATED" "OBSOLETE"))
  ;; (current-block-header nil
  ;;  ;; precomputed
  ;;  :type (or null string))
  )

;; An idea: we might have several contexts for a file,
;; each corresponding to a position,
;; within an editing session of the file;
;; when some (simple) operations on a file are performed,
;; positions can be recomputed
;; so that contexts can be recomputed without performing complete parsing.
;; We only need to preserve those contexts which change between consequent blocks.
;; I.e., conseqent blocks found in different Org sections but on the same level
;; may all be serviced (e.g., evaluated correctly) interactively
;; by maintaining a single context for them.
;; Anyway, for now, we don't even have anything for editing sessions.

(eval-when (:compile-toplevel :load-toplevel)
  (defun get-normalized-lang-name (block-lang &optional context)
    (declare (ignore
              ;; We might need Org buffer's context to normalize a lang name in some cases
              ;; but for now I haven't found such cases
              context))
    (cond
      ((string= "lisp" block-lang)
       "common-lisp")
      ((or (string= "elisp" block-lang)
           (string= "emacs-lisp" block-lang))
       "emacs-lisp"))))

(eval-when (:compile-toplevel :load-toplevel)
  (defun lang-equal (string1 string2)
    (declare (type string string1 string2))
    (string= (get-normalized-lang-name string1)
             (get-normalized-lang-name string2))))

(defmacro with-context (context-place &body body)
  `(symbol-macrolet ((now-reading (context-now-reading ,context-place))
                     (current-level (context-current-level ,context-place))
                     (header-args-stack (context-header-args-stack ,context-place))
                     (langwise-header-args-stacks (context-langwise-header-args-stacks ,context-place))
                     (todo-keywords-stack (context-todo-keywords-stack ,context-place))
                     ;; (block-headers-stack (context-block-headers-stack ,context-place))
                     (left-whitespace-p (context-left-whitespace-p ,context-place))
                     (left-whitespace-column (context-left-whitespace-column ,context-place))
                     (current-header-args (context-current-header-args ,context-place))
                     (current-langwise-header-args (context-current-langwise-header-args ,context-place))
                     (current-todo-keyword (context-current-todo-keyword ,context-place))
                     (current-todo-keywords-labelling-ignored-content (context-current-todo-keywords-labelling-ignored-content ,context-place))
                     ;; (current-block-header (context-current-block-header ,context-place))
                     )
     (macrolet ((langwise-header-args-stack (lang)
                  `(alist-get ,lang langwise-header-args-stacks
                              nil nil #'lang-equal))
                (current-langwise-header-args (lang)
                  `(alist-get ,lang current-langwise-header-args
                              nil nil #'lang-equal)))
       ,@body)))

(defmacro update-header-args-f (header-args new-header-args)
  (let ((old (make-symbol "OLD-HEADER-ARGS"))
        (key (make-symbol "KEY"))
        (value (make-symbol "VALUE")))
    `(let ((,old ,header-args))
       ;; Note: this way, properties listed later in new-header-args
       ;; will take priority over those listed earlier
       ;; I don't know what Org says about duplicate properties;
       ;; I just presume it UB for now.
       (loop for (,key . ,value) in ,new-header-args
             do (aif (assoc ,key ,old :test #'eq)
                    (setf (cdr it) ,value)
                  (push (cons ,key ,value) ,header-args)))
       ,header-args)))

(defun merge-header-args (context header-args)
  "header-args are presumed to be on the current level of context.

Return value is unspecified."
  (when header-args
    (with-context context
      (push (cons current-level header-args) header-args-stack)
      (update-header-args-f current-header-args header-args)
      (loop for cons in current-langwise-header-args
            do (update-header-args-f (cdr cons) header-args)))))

(defun compute-header-args (header-args-stack)
  (loop with header-args
        for (_level . alist) in header-args-stack
        do (update-header-args-f header-args alist)
        finally (return header-args)))

(defmacro with-one-time-header-args (header-args context &body body)
  ;; TODO: Maybe just update current-header-args, without updating the stack
  `(let (,(o-o header-args) ,(o-o context))
     ;; TODO: Maybe it makes sense to allow expressions in header-args
     ;; to see the scope of context.
     ;; But, at the very least, the order of arguments better be
     ;; “context header-args” then.
     (with-context ,context
       (merge-header-args ,context ,header-args)
       (prog1 (locally ,@body)
         (when ,header-args
           (pop header-args-stack)
           (setf current-header-args
                 (compute-header-args header-args-stack)))))))

(defun merge-langwise-header-args (lang context header-args)
  "header-args are presumed to be on the current level of context.

Return value is unspecified."
  (declare (type string lang))
  (when header-args
    (with-context context
      (push (cons current-level header-args)
            (langwise-header-args-stack lang))
      (update-header-args-f (current-langwise-header-args lang)
                            header-args))))

(defmacro with-one-time-langwise-header-args ( lang header-args context
                                               &body body)
  ;; TODO: Maybe just update current-langwise-header-args, without updating the stack
  `(let (,(o-o lang) ,(o-o header-args) ,(o-o context))
     ;; TODO: Maybe it makes sense to allow expressions in header-args
     ;; to see the scope of context.
     ;; But, at the very least, the order of arguments better be
     ;; “context header-args” then.
     (with-context ,context
       (merge-langwise-header-args ,lang ,context ,header-args)
       (prog1 (locally ,@body)
         (when ,header-args
           (pop (langwise-header-args-stack ,lang))
           (setf (current-langwise-header-args ,lang)
                 (compute-header-args
                  (langwise-header-args-stack ,lang))))))))

(defun compute-todo-keyword (todo-keywords-stack)
  (cdar todo-keywords-stack))

(defun make-empty-context ()
  (make-context))

(defvar all-contexts nil
  "Alist: keys are streams, values are contexts")
#+akater-types
(declaim (type (alist stream context) all-contexts))

(defun delete-all-contexts ()
  (nullf all-contexts)
  'all-contexts-deleted)

;; Due to how asdf works, we'll need to associate contexts to files too

(defun initialize-context (stream)
  (aprog1 (make-empty-context)
    (push (cons stream it) all-contexts)))

(defun ensure-context (stream)
  (or (aand (assoc stream all-contexts :test #'eq)
            (cdr it))
      (initialize-context stream)))

(defun get-stream (context)
  (declare (type context context))
  (car (rassoc context all-contexts :test #'eq)))

(defmacro with-context-of-stream (stream &body body)
  (let ((s (make-symbol "STREAM-WITH-CONTEXT")))
    `(let* ((,s ,stream)
            (context (ensure-context ,s)))
       (flet ((delete-context ()
                ;; todo: assert the stream was deleted
                (setq all-contexts
                      (delete ,s all-contexts :count 1
                                              :key #'car :test #'eq))))
         (with-context context
           ;; TODO: maybe (macrolet ((with-context ( &whole form
           ;;                                        place &body body)
           ;;                          (if (eq 'context place) `(progn ,@body)
           ;;                            form)))
           ;;               ,@body)
           ,@body)))))

(unless (find-package "ORG-SYNTAX-TOPLEVEL")
  ;; This is the package where most symbols read in org headers will go
  ;; Maybe I'd rather read them uninterned
  ;; but there doesn't seem to be an easy way to.
  (let ((*package*
          (apply #'make-package "ORG-SYNTAX-TOPLEVEL"
                 #+sbcl '(:internal-symbols 53
                          :external-symbols 0)
                 #-sbcl nil)))
    (import '(t nil))))

(defun error-in-stream-at-position (stream datum &rest arguments)
  ;; todo: try to define this via define-condition
  (apply #'error (concatenate 'string datum " in ~S at position ~S")
         (append arguments (list stream (file-position stream)))))

(defun warn-in-stream-at-position (stream datum &rest arguments)
  ;; todo: try to define this via define-condition
  (apply #'warn (concatenate 'string datum " in ~S at position ~S")
         (append arguments (list stream (file-position stream)))))

(defvar cl-readtable cl:*readtable*
  ;; There is no standard way to obtain the default readtable
  ;; so we just hope it's that.
  )

(defun parse-header-args (string &optional (start 0) &aux header-args)
  (declare (type list header-args)
           (type string string)
           (type index start))
  (with-open-stream (stream (make-string-input-stream string start))
    (let ((*readtable*
            ;; not really, var= won't work, but that's it for now
            cl-readtable)
          (*package* (find-package "ORG-SYNTAX-TOPLEVEL"))
          tokens-so-far (end-token (make-symbol "END-OF-HEADER-ARGS-LINE")))
      (loop
        (let ((token (read stream nil end-token t)))
          (flet ((push-header-arg ()
                   (when tokens-so-far
                     (nreversef tokens-so-far)
                     (let (neverp sometimesp)
                       (case (car tokens-so-far)
                         ((:eval-when)
                          (loop for cell on (cdr tokens-so-far)
                                do (setf (car cell)
                                         (let ((it (car cell)))
                                           (check-type it string-designator)
                                           (if (or (string= 'nil it)
                                                   (string= 'never it))
                                               (progn (setf neverp t) it)
                                             (setf sometimesp
                                                   (cond
                                                     ((string= 'compile it)
                                                      :compile-toplevel)
                                                     ((string= 'load it)
                                                      :load-toplevel)
                                                     ((string= 'execute it)
                                                      :execute)
                                                     (:else it)))))))
                        
                          (when (and neverp sometimesp)
                            (warn-in-stream-at-position stream "Ambiguous :eval-when spec: nil or never, and something else,"))
                          (when neverp (nullf (cdr tokens-so-far))))))
                     (push (nullf1 tokens-so-far) header-args))))
            (cond
              ((eq end-token token) (push-header-arg)
               (return header-args))
              ((keywordp token) (push-header-arg)
               (case token
                 (:var
                  ;; todo: set readtable
                  ;; to one understanding infix =
                  )))))
          (push token tokens-so-far))))))

;; (parse-header-args "  :tangle no :eval-when compile load :keep-expected t  ")

(defvar *allow-all-stars-lines* 'warn
  "If set to nil, error, if set to 'warn, warn.  Otherwise, silently accept.")

(defun generic-char-reader (stream char)
  (declare (type (not (member #\Newline #\#)) char))
  (with-context-of-stream stream
    (when left-whitespace-p
      (if (whitespace-char-p char)
          (incf left-whitespace-column)
          (nullf left-whitespace-p)))
    (unless (peek-char nil stream nil nil t)
      ;; Note: if files are always ended with newline, this is not needed.
      (delete-context)))
  (values))

;; it would be better to define the following two
;; via generic-char-reader, inlined, so that its checks are simplified

(defun whitespace-char-reader (stream char)
  (declare (type (member #\Space #\Tab) char))
  (declare (ignore char))
  (with-context-of-stream stream
    (when left-whitespace-p
      (incf left-whitespace-column))
    (unless (peek-char nil stream nil nil t)
      ;; Note: if files are always ended with newline, this is not needed.
      (delete-context)))
  (values))

(defun non-whitespace-char-reader (stream char)
  (declare (type (not (member #\Space #\Tab)) char))
  (declare (ignore char))
  (with-context-of-stream stream
    (when left-whitespace-p
      (nullf left-whitespace-p))
    (unless (peek-char nil stream nil nil t)
      ;; Note: if files are always ended with newline, this is not needed.
      (delete-context)))
  (values))

(defun newline-reader (stream char)
  (declare (ignore char))
  (with-context-of-stream stream
    (psetf left-whitespace-p t
           left-whitespace-column 0)
    (case now-reading
      ;; todo: better do this with colon-reader
      (properties
       (multiple-value-bind (line missing-newline-p) (read-line stream
                                                                nil nil t)
         (if (or missing-newline-p (null line))
             (progn (warn-in-stream-at-position stream "Unclosed :properties:")
                    (delete-context))
           (progn
             (unread-char #\Newline stream)
             (cond
               ((string-prefix-p ":header-args:" line)
                (cond
                  ((= (length ":header-args:")
                      (length line))
                   (error-in-stream-at-position stream "Can't parse header-args property line ~S"
                                                line))
                  ((not (whitespace-char-p (char line (length ":header-args:"))))
                   (let* ((lang-start (length ":header-args:"))
                          (lang-end (position #\: line :start lang-start :test #'char=))
                          (lang (progn
                                  (unless lang-end
                                    (error-in-stream-at-position stream "Incorrect syntax after :header-args: on line ~S: no closing colon;"
                                                                 line))
                                  (subseq line lang-start lang-end)))
                          (prefix-end (1+ lang-end)))
                     (cond
                       ((= prefix-end (length line))
                        (warn-in-stream-at-position stream "Empty properties for ~S" line)
                        nil)
                       ((not (whitespace-char-p (char line prefix-end)))
                        (error-in-stream-at-position stream "Can't parse header-args property line ~S: not a whitespace after ~S;"
                                                     line (subseq line 0 prefix-end)))
                       ((merge-langwise-header-args lang context (parse-header-args line prefix-end))))))
                  ((merge-header-args
                    context
                    (parse-header-args line (length ":header-args:"))))))
               ((string-equal ":end:" line)
                (nullf now-reading)))))))
      (otherwise
       (let* ((end (make-symbol "END-OF-ORG-STREAM"))
              (next (peek-char t stream nil end t)))
         (cond
           ((eq end next)
            (delete-context))
           ((char= #\* next)
            ;; I can't find a good verb to describe what's happenging
            ;; This is part of the reason why I haven't isolated it.
            (let (todo-keyword)
              (when-let ((level
                          (multiple-value-bind (line missing-newline-p) (read-line stream t nil t)
                            (if missing-newline-p (delete-context)
                              (unread-char #\Newline stream))
                            (let ((it (position-not #\* line)))
                              (cond
                                ((null it) (if missing-newline-p (progn (warn "All stars on the last line") nil)
                                             (case *allow-all-stars-lines*
                                               (warn
                                                (warn-in-stream-at-position stream "Probably missing whitespace after *..")
                                                (length line))
                                               ((nil)
                                                (error-in-stream-at-position stream "Missing whitespace after *.."))
                                               (otherwise (length line)))))
                                (missing-newline-p (let ((headlinep (char= #\Space (char line it))))
                                                     (when headlinep
                                                       (warn "Headline on the last line."))
                                                     (if headlinep
                                                         (warn "Ignoring headline on the last line.")
                                                       (warn "Ignoring last line starting with “*”.")))
                                                   nil)
                                (:else
                                 (when-debug
                                   (format t "Reading headline ~S at position ~S~&"
                                           line (file-position stream)))
                                 (cond
                                   ((char= #\Space (char line it))
                                    (when-let ((maybe-todo-keyword
                                                (when-let ((maybe-keyword-start
                                                            (position-not #\Space line :start (1+ it))))
                                                  (subseq line maybe-keyword-start (position #\Space line :start (1+ maybe-keyword-start))))))
                                      (setq todo-keyword
                                            ;; todo: check if maybe-todo-keyword is in list of declared todo-keywords
                                            (or (find maybe-todo-keyword '("TEST-PASSED" "TEST-FAILED" "TEST") :test #'string=)
                                                (find maybe-todo-keyword current-todo-keywords-labelling-ignored-content :test #'string=))))
                                    it)
                                   (:else
                                    ;; TODO: Look into the line better, write more to the output
                                    (unless (position #\* line :start (1+ it))
                                      (warn-in-stream-at-position stream "Likely missing whitespace in line ~S"
                                                                  line))
                                    (warn-in-stream-at-position stream "Skipping possible headline ~S"
                                                                line)
                                    nil))))))))
                (setf current-level level)
                (let ((global-dropped (symbol-macrolet ((obsolete-header-args
                                          (and header-args-stack
                                               (<= level (caar header-args-stack)))))
                         (let ((dropped obsolete-header-args))
                           (loop while obsolete-header-args
                                 do (pop header-args-stack))
                           (when1 dropped
                             (setf current-header-args
                                   (compute-header-args header-args-stack)))))))
                 (loop for (lang . header-args-stack) in langwise-header-args-stacks
                       do (symbol-macrolet ((obsolete-header-args
                                             (and header-args-stack
                                                  (<= level (caar header-args-stack)))))
                            (let ((dropped obsolete-header-args))
                              (loop while obsolete-header-args
                                    do (pop header-args-stack))
                              (when (or dropped global-dropped)
                                (setf (current-langwise-header-args lang)
                                      (compute-header-args header-args-stack)))))))
                (symbol-macrolet ((obsolete-todo-keywords
                                   (and todo-keywords-stack
                                        (<= level (caar todo-keywords-stack)))))
                  (let ((dropped obsolete-todo-keywords))
                    (loop while obsolete-todo-keywords
                          do (pop todo-keywords-stack))
                    (when-debug
                      (format t "Context before todo-keyword check:~%~S~&" context))
                    (when todo-keyword 
                      (push (cons current-level todo-keyword)
                            todo-keywords-stack))
                    (when (or dropped todo-keyword)
                      (setf current-todo-keyword
                            (compute-todo-keyword todo-keywords-stack))))))))
           ((char= #\: next)
            (multiple-value-bind (line missing-newline-p) (read-line stream nil nil t)
              (if missing-newline-p (delete-context)
                (unread-char #\Newline stream))
              (cond ((string-equal ":properties:" line)
                     (if missing-newline-p (warn-in-stream-at-position stream "~S on the last line"
                                                                       line)
                       (setf now-reading 'properties)))
                    ((string-equal ":end:" line)
                     (error-in-stream-at-position stream "Can't match closing substring ~S with anything"
                                                  line))))))))))
  (values))

(defun sharp-plus-end_-line-p (string)
  (declare (type string string))
  (the (or null index)
       (loop for i of-type index from 0 for c across string
             do (unless (whitespace-char-p c)
                  (return (when (and (>= (- (length string) i)
                                         (;; block type must be non-empty
                                          1+ (length "#+end_")))
                                     (string-equal "#+end_"
                                              (subseq string i
                                                      (+ i (length "#+end_")))))
                            i))))))
;; (sharp-plus-end_-line-p "   #+end_src")
;; (sharp-plus-end_-line-p "   #+end_src  ")
;; (sharp-plus-end_-line-p "#+end_src  ")
;; (sharp-plus-end_-line-p "#+end_example")
;; (sharp-plus-end_-line-p "#+end")
;; (sharp-plus-end_-line-p "#+end_")
;; (sharp-plus-end_-line-p "#+end_t")

;; (with-input-from-string (s "#+begin_src lisp :eval-when compile load
;; (wow)
;; #+end_src
;; ")
;;   (let ((*readtable* org-syntax:readtable))
;;     (read s)))

(defun macroexpander (macroexp-spec &optional (package-designator
                                               "COMMON-LISP"))
  "Return type: valid entry for a car in a compound form.

MACROEXP-SPEC is presumed to have prefix  macroexp ."
  (declare (type string macroexp-spec))
  (let ((suffix (subseq macroexp-spec (length "macroexp"))))
    (or (when (zerop (length suffix))
          'macroexpand-1)
        (multiple-value-bind (number-suffix end) (ignore-errors
                                                  (read-from-string suffix))
          (when (and end (integerp end) (integerp number-suffix)
                     (= end (length suffix)))
            (if (eql 1 number-suffix)
                'macroexpand-1
                `(lambda (form) (nest ,number-suffix #'macroexpand-1 form)))))
        (find-symbol macroexp-spec (find-package package-designator))
        (intern macroexp-spec (find-package package-designator)))))

(defun get-block-specific-header-args (context block-lang)
  (with-context context
    (let ((header-args (copy-tree current-header-args)))
      (loop for (key . value) in (current-langwise-header-args block-lang)
            do (setf (alist-get key header-args nil nil #'eq)
                     value))
      header-args)))

(defgeneric skip-block-p (lang context
                          &key header-args
                               ;; for caching purposes only, as we can't compute it from lang object and context alone
                               ;; we might add a label slot, though, and ensure lang object with that label at each block
                               )
  ;; we might want to mention the block (its precise block-lang, or its position)
  ;; in debug messages.
  ;; The relevant data is in context,
  ;; hence the context is used as argument.
  ;; todo: check what argument order is best for precedence
  (:method-combination or))

(defmethod skip-block-p or ((lang lang) (context context) &key header-args)
  (declare (ignore lang context header-args)))

(defun skip-block-p/plain (block-lang context &key (header-args (get-block-specific-header-args context block-lang))
                                                   (normalized-lang (get-normalized-lang-name block-lang)))
  (skip-block-p (org-syntax-language:ensure normalized-lang :block-lang block-lang) context :header-args header-args))

(defgeneric get-readtable (lang context
                           &key header-args
                                ;; for caching purposes only, as we can't compute it from lang object and context alone
                                ;; we might add a label slot, though, and ensure lang object with that label at each block
                                )
  ;; context holds langwise header-args and global header-args separately
  ;; we might want to mention the block (its precise block-lang, or its position)
  ;; in debug messages.
  ;; The relevant data is in context,
  ;; hence the context is used as argument.
  ;; todo: check what argument order is best for precedence
  (:method-combination or))

(defun get-readtable/plain (block-lang context &key (header-args (get-block-specific-header-args context block-lang))
                                                    (normalized-lang (get-normalized-lang-name block-lang)))
  (get-readtable (org-syntax-language:ensure normalized-lang :block-lang block-lang) context :header-args header-args))

(defgeneric get-package (lang context
                         &key header-args
                              ;; for caching purposes only, as we can't compute it from lang object and context alone
                              ;; we might add a label slot, though, and ensure lang object with that label at each block
                              )
  ;; context holds langwise header-args and global header-args separately
  ;; we might want to mention the block (its precise block-lang, or its position)
  ;; in debug messages.
  ;; The relevant data is in context,
  ;; hence the context is used as argument.
  ;; todo: check what argument order is best for precedence
  (:method-combination or))

(defun get-package/plain (block-lang context &key (header-args (get-block-specific-header-args context block-lang))
                                                  (normalized-lang (get-normalized-lang-name block-lang)))
  (get-package (org-syntax-language:ensure normalized-lang :block-lang block-lang) context :header-args header-args))

;; todo: we need to set *package* during compilation, because defstruct forms intern symbols into it at compile time
;; we will probably have to produce (in-package ..) forms for each instance of :package
;; it would be nice if didn't produce them excessively, namely, restoring on each step up the org tree
;; but i'm not sure this can be avoided

(defun sharp-reader (stream char)
  (declare (ignore char))
  (with-context-of-stream stream
    (when left-whitespace-p
      (let ((end (make-symbol "END-OF-ORG-STREAM"))
            ;; todo: whenever end is returned, delete the stream from all-contexts
            ;; this should be done in with-context-of-stream
            )
        (multiple-value-bind (line missing-newline-p) (read-line stream nil end t)
          (cond
            (missing-newline-p
             (warn-in-stream-at-position stream "Ignoring last line starting with “#”")
             (delete-context)
             (values))
            ((eq end line)
             (delete-context)
             (values))
            ;; todo: maybe support reading first line
            ;; todo: maybe support description (subtitle), title and so on
            ;; todo: maybe support TODO sequences and so on
            ((string-prefix-p "+property:" line)
             ;; todo: support this
             (unless (= 0 left-whitespace-column)
               (warn-in-stream-at-position stream "Suspicious placement of “#+property”: non-zero indent;"))
             (let ((some-keyword-start
                     (position-not #\Space line
                                   :start (length "+property:"))))
               (when some-keyword-start
                 (let* ((some-keyword-end
                          (position #\Space line :start some-keyword-start))
                        (some-keyword
                          (subseq line some-keyword-start some-keyword-end)))
                   (cond
                     ((string-equal "header-args" some-keyword)
                      (when (aand header-args-stack
                                  (assoc 0 it))
                        (warn-in-stream-at-position stream "Looking at “#+property header-args ...” line while header-args-stack has a record at level 0,"))
                      (merge-header-args
                       context
                       (parse-header-args line (or some-keyword-end
                                                   (length line)))))
                     ((and (string-prefix-p "header-args" some-keyword)
                           (char= #\: (char some-keyword
                                            (length "header-args"))))
                      (let ((lang
                              (subseq some-keyword
                                      (1+ (length "header-args")))))
                        (when (aand (alist-get lang langwise-header-args-stacks
                                               nil nil #'lang-equal)
                                    (assoc 0 it))
                          (warn-in-stream-at-position stream "Looking at “#+property header-args:<lang> ...” line while langwise-header-args-stack has a record at level 0"))
                        (merge-langwise-header-args
                         lang
                         context
                         (parse-header-args line (or some-keyword-end
                                                     (length line))))))
                     (:else (warn-in-stream-at-position stream "Can't parse “#+property:” line"))))))
             (unread-char #\Newline stream)
             nil)
            ((string-prefix-p "+begin_" line)
             (when-debug
               (format t "Context at block header ~S:~%~S~&"
                       (subseq line 1) (plist-with-keys context)))
             (let ((block-header (subseq line (length "+begin_")))
                   type-end)
               (declare (ignore block-header))
               (let* ((block-type (subseq line (length "+begin_")
                                          (setq type-end
                                                (position #\Space line
                                                          :start (length "+begin_")))))
                      (no-lang (null type-end))
                      (lang-start (when type-end
                                    (1+
                                     ;; we simply presume there's a single space between type and lang
                                     type-end)))
                      (lang-end lang-start)
                      (block-lang (when lang-start
                                    (subseq line lang-start
                                            (setq lang-end
                                                  (position #\Space line
                                                            :start lang-start))))))
                 (if (string-equal "src" block-type)
                     (progn
                       (when no-lang (error "No lang specified for ~S block" block-type))
                       (unless (and todo-keywords-stack
                                    (= current-level (caar todo-keywords-stack)))
                         (if (not (or (string= "lisp" block-lang)
                                      (string= "elisp" block-lang)
                                      (string= "emacs-lisp" block-lang)))
                             (error "Don't know how to read ~S block" block-lang)
                             (with-one-time-langwise-header-args block-lang (parse-header-args line (or lang-end
                                                                                                        (length line))) context
                               (when-debug
                                 (format t "Context within block ~S:~%~S~&"
                                         (subseq line 1) (plist-with-keys context)))
                               ;; todo: do not read test blocks
                               (let* ((length 0)
                                      ;; todo: maybe use streams instead of collecting substrings
                                      (block-content-lines
                                        (let (suspicious-indentation)
                                          (prog1 (loop with end = (make-symbol "END-OF-ORG-STREAM")
                                                       with end_-indent = 0
                                                       ;; with block-terminated
                                                       with line = ""
                                                       do (cond
                                                            ((eq end
                                                                 (setq line
                                                                       (read-line stream
                                                                                  nil end t)))
                                                             (delete-context)
                                                             (error "End of file but src block is not terminated")
                                                             (loop-finish))
                                                            ((setq end_-indent
                                                                   (sharp-plus-end_-line-p
                                                                    line))
                                                             (let* (type-end
                                                                    (block-type/end
                                                                      (subseq line (+ end_-indent (length "#+end_"))
                                                                              (setq type-end
                                                                                    (position #\Space line
                                                                                              :start (+ end_-indent (length "#+end_")))))))
                                                               (cond
                                                                 ((string-equal "src" block-type/end)
                                                                  (when (and type-end (>= (length line) type-end))
                                                                    (if (string-suffix-whitespace-p type-end line)
                                                                        (warn-in-stream-at-position stream "Trailing whitespace after #+end_src")
                                                                      (warn-in-stream-at-position stream "Garbage after #+end_src")))
                                                                  (unless (= left-whitespace-column
                                                                             end_-indent)
                                                                    (warn-in-stream-at-position stream "Indentation mismatch for #+begin_ .. #+end_"))
                                                                  (nullf left-whitespace-p)
                                                                  (unread-char #\Newline stream)
                                                                  (loop-finish))
                                                                 (:else
                                                                  (when (= left-whitespace-column
                                                                           end_-indent)
                                                                    (warn-in-stream-at-position stream "Suspicious indentation of “#+end_~A” after #+begin_src"
                                                                          block-type/end))
                                                                  (when (< end_-indent left-whitespace-column)
                                                                    (warn-in-stream-at-position stream "Very suspicious indentation of “#+end_~A” after #+begin_src"
                                                                          block-type/end))))))
                                                            (:else
                                                             (incf length (1+ (length line)))
                                                             (unless (string-prefix-whitespace-p
                                                                      left-whitespace-column line)
                                                               (setf suspicious-indentation t))))
                                                          ;; todo: do not collect lines if :read no
                                                       collect line)
                                            (when suspicious-indentation
                                              (warn-in-stream-at-position stream "Suspicious indentation within #+begin_ .. #+end_ block: content “outside” of a block;"))))))

                                 (block read-form
                                   (let ((header-args
                                           ;; TODO: What about one-time-langwise-header-args? afaicr, it temporarily pushes langwise header args but does not compute header args for block
                                           ;; as I recomputed them here.
                                           ;; I suspect I should simply compute them in ..-one-time-..
                                           ;; But I don't want to put block-lang into context for some reason (maybe because the mode within a block is essentially different and thus context should be essentially different)
                                           ;; while having block-lang passed to generics skip-block, get-readtable, get-package is somewhat useful.
                                           (get-block-specific-header-args context block-lang)))
                                     (when-debug
                                       (format t "Actual header args in block: ~S~&"
                                               header-args))
                                     (when (or (aand (assoc :disabled header-args :test #'eq)
                                                     (progn
                                                       (when (cddr it)
                                                         (error-in-stream-at-position stream "Unsupported syntax for ~S in ~S"
                                                                :disabled header-args))
                                                       (pop it) (carf it)
                                                       (and (symbolp it)
                                                            (string= 't it))))
                                               (aand (assoc :read header-args :test #'eq)
                                                     (progn
                                                       (when (cddr it)
                                                         (error-in-stream-at-position stream "Unsupported syntax for ~S in ~S"
                                                                :read header-args))
                                                       (pop it) (carf it)
                                                       (and (symbolp it)
                                                            (string= 'no it)))))
                                       (return-from read-form))
                                     (let ((normalized-lang (get-normalized-lang-name block-lang)))
                                       (assert (or
                                                ;; TODO: replace this “or” form with something more appropriate
                                                (string= "common-lisp" normalized-lang)
                                                (string= "emacs-lisp" normalized-lang)))
                                       (when (skip-block-p/plain block-lang context :header-args header-args :normalized-lang normalized-lang)
                                         (return-from read-form))
                                       (let ((block-content
                                              (let ((out (make-string
                                                          length
                                                          :initial-element #\Space)))
                                                (loop for line in block-content-lines
                                                      with i of-type index = 0
                                                      do (loop for c across line
                                                               do (setf (char out (incf1 i))
                                                                        c))
                                                         (setf (char out (incf1 i))
                                                               #\Newline))
                                                out))
                                             (*readtable* (or
                                                           ;; TODO: maybe implement this header-args lookup in a method
                                                           ;; but it has to always come first;
                                                           ;; so, if we implement this in a method,
                                                           ;; should we specify get-readtable's method as most-specific-last?
                                                           ;; I might define context-with-readtable but I don't feel like it so far
                                                           (aand (assoc :named-readtable header-args :test #'eq)
                                                                 (progn
                                                                   (when (assoc :readtable header-args :test #'eq)
                                                                     (error-in-stream-at-position stream "~S and ~S simultaneously in header args ~S"
                                                                                                  :named-readtable :readtable header-args))
                                                                   (when (cddr it)
                                                                     (error-in-stream-at-position stream "Unsupported syntax for ~S in ~S"
                                                                                                  :named-readtable header-args))
                                                                   (pop it) (carf it)
                                                                   (and (or (symbolp it) (error-in-stream-at-position stream "~S syntax only supports symbols, not ~S"
                                                                                                                      :named-readtable it))
                                                                        (aprog1 (eval
                                                                                 `(,(or (find-symbol "ENSURE-READTABLE"
                                                                                                     (or (find-package "NAMED-READTABLES")
                                                                                                         (error-in-stream-at-position stream "~S is used in header args ~S but package ~A does not exist"
                                                                                                                                      :named-readtable header-args "NAMED-READTABLES")))
                                                                                        (error "Symbol ~A not found in package ~A"
                                                                                               "ENSURE-READTABLE" "NAMED-READTABLES"))
                                                                                   ',it))
                                                                          (unless (readtablep it)
                                                                            (error-in-stream-at-position stream "Not a readtable specified in header args ~S"
                                                                                                         header-args))))))
                                                           (aand (assoc :readtable header-args :test #'eq)
                                                                 (progn
                                                                   (when (cddr it)
                                                                     (error-in-stream-at-position stream "Unsupported syntax for ~S in ~S"
                                                                                                  :readtable header-args))
                                                                   (pop it) (carf it)
                                                                   (and (or (symbolp it) (error-in-stream-at-position stream "~S syntax only supports symbols, not ~S"
                                                                                                                      :readtable it))
                                                                        (aprog1 (symbol-value it)
                                                                                (unless (readtablep it)
                                                                                  (error-in-stream-at-position stream "Not a readtable specified in header args ~S"
                                                                                                               header-args))))))
                                                           (get-readtable/plain block-lang context :header-args header-args :normalized-lang normalized-lang)))
                                             (*package* (get-package/plain block-lang context :header-args header-args :normalized-lang normalized-lang)))
                                         (let ((forms
                                                (with-input-from-string (in block-content)
                                                  (let ((end (make-symbol "END-OF-ORG-BLOCK-CONTENT")) expr)
                                                    ;; we say “expr” but it should probably
                                                    ;; always be considered a top-level form
                                                    ;; or a declaration
                                                    (loop until (eq end (setq expr
                                                                              (read in nil
                                                                                    end t)))
                                                          collect expr)))))
                                           (let (custom-wrapper results-spec)
                                             (cond
                                               ((setq custom-wrapper
                                                      (assoc :eval-when header-args :test #'eq))
                                                ;; Maybe, if lang-elisp is non-nil
                                                ;; we should instead wrap in #EL::eval-when-compile
                                                ;; or in a different relevant header.
                                                `(eval-when ,(cdr custom-wrapper) ,@forms))
                                               ((setq results-spec (assoc :results header-args :test #'eq))
                                                ;; Currently, this is almost entirely useless because we skip test blocks.
                                                ;; TODO: Make this much more useful
                                                ;; TODO: Port all my custom headers for elisp blocks
                                                (let ((macroexp-spec (let ((string ""))
                                                                       (declare (type string string))
                                                                       (when (member "macroexp" results-spec
                                                                                     :test #'string-prefix-p
                                                                                     :key (lambda (entry)
                                                                                            (setq string (ensure-string entry))))
                                                                         string))))
                                                  (if macroexp-spec (when forms
                                                                      (let* (last
                                                                             (most (loop for rest on forms
                                                                                         if (cdr rest) collect (car rest)
                                                                                         else do (setq last (car rest)))))
                                                                        (wrap-if most `(progn
                                                                                         ,@most
                                                                                         ,last)
                                                                                 last
                                                                                 `(,(macroexpander macroexp-spec)
                                                                                   ',last))))
                                                      (if (null (cdr forms)) (car forms) `(progn ,@forms)))))
                                               ((null (cdr forms)) (car forms))
                                               (`(progn ,@forms))))))))))))))
                     (let (suspicious-indentation)
                       (when no-lang
                         (unless (member block-type '("quote" "comment" "verse" "center" "verbatim")
                                         :test #'string-equal)
                           (error "No lang specified for ~S block" block-type)))
                       (prog1 (loop with end = (make-symbol "END-OF-ORG-STREAM")
                                    with end_-indent = 0
                                    ;; with block-terminated
                                    with line = ""
                                    do (cond
                                         ((eq end
                                              (setq line
                                                    (read-line stream
                                                               nil end t)))
                                          (delete-context)
                                          (error "End of file but src block is not terminated")
                                          (loop-finish))
                                         ((setq end_-indent
                                                (sharp-plus-end_-line-p
                                                 line))
                                          (let* (type-end
                                                 (block-type/end
                                                   (subseq line (+ end_-indent (length "#+end_"))
                                                           (setq type-end
                                                                 (position #\Space line
                                                                           :start (+ end_-indent (length "#+end_")))))))
                                            (cond
                                              ((string-equal block-type block-type/end)
                                               (when (and type-end (>= (length line) type-end))
                                                 (if (string-suffix-whitespace-p type-end line)
                                                     (warn-in-stream-at-position stream "Trailing whitespace after #+end_~A"
                                                                                 block-type)
                                                   (warn-in-stream-at-position stream "Garbage after #+end_src")))
                                               (unless (= left-whitespace-column
                                                          end_-indent)
                                                 (warn-in-stream-at-position stream "Indentation mismatch for #+begin_ .. #+end_"))
                                               (nullf left-whitespace-p)
                                               (loop-finish))
                                              (:else
                                               (when (= left-whitespace-column
                                                        end_-indent)
                                                 (warn-in-stream-at-position stream "Suspicious indentation of “#+” within #+begin_ .. #+end_ block"))))))
                                         (:else
                                          (unless (string-prefix-whitespace-p
                                                   left-whitespace-column line)
                                            (setf suspicious-indentation t))))
                                    collect line)
                         (when suspicious-indentation
                           (warn-in-stream-at-position stream "Suspicious indentation within #+begin_ .. #+end_ block: content “outside” of a block;")))
                       nil)
                     ;; (push (cons left-whitespace-column block-header)
                     ;;       block-headers-stack)
                     ))))))))))

(defparameter org-syntax:readtable
  (let ((*readtable* empty-readtable))
    (loop for c from 0 below char-code-limit
          do (set-syntax-from-char (code-char c) #\Space)
             (set-macro-character (code-char c) #'generic-char-reader t))
    (set-macro-character #\# #'sharp-reader t)
    (set-macro-character #\Newline #'newline-reader t)
    *readtable*))

;; org-syntax:with-org-syntax is ugly

(defmacro org-syntax:with (&body body)
  ;; This does not read well even with prefix
  `(let ((*readtable* org-syntax:readtable)
         (*package* *package*))
     ,@body))

(setf (macro-function 'with-org-syntax)
      (macro-function 'org-syntax:with))

#+akater-types
(org-syntax.system::fbind-list-of-checkers)

;; These are nice looking but require some external “convention package”:

;; syntax:with-org
;; with-syntax:org
;; syntax:org

;; (defmacro org-syntax:enabled (&body body)
;;   ;; The only issue with this name is,
;;   ;; Lisp itself will often print it without the package name
;;   ;; and this is bad
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; ;; The following kinda read well both with and without the package name

;; (defmacro org-syntax:reading (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; (defmacro org-syntax:# (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; (defmacro org-syntax:covered (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; (defmacro org-syntax:ongoing (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; (defmacro org-syntax:loaded (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))

;; (defmacro org-syntax:bound (&body body)
;;   `(let ((*readtable* org-syntax:readtable)
;;          (*package* *package*))
;;      ,@body))


;; imitating swank/source-path-parser::skip-toplevel-forms but with our readtable
;; this errors, bad
;; (org-syntax:with-org-syntax
;;   (with-input-from-string (stream "
;; #+begin_src lisp
;; ()
;; #+end_src

;; #+begin_src lisp
;; ()
;; #+end_src

;; #+begin_src lisp
;; (wow)
;; #+end_src
;; ") (let ((*read-suppress* t) f)
;;      (setq f (read stream))
;;      ;; if uncommented, this form errors
;;      ;; (setq f (read stream))
;;      (peek-char t stream nil nil)
;;      f)))
