(in-package #:org-syntax.util)

#+akater-types
(eval-when (:load-toplevel)
  (loop for (name . lambda-expr) in org-syntax.system::type-checkers/list-of
        do (unless (fboundp name) (format t "Compiling ~S~&" name) (compile name lambda-expr))))

(defmacro aprog1 (x &body body)
  `(let ((it ,x)) (prog1 it ,@body)))

;; note: in swank/source-path-parser,
;; skip-whitespace (stream) is implemented as just (peek-char t stream nil nil)

(defun string-prefix-p (prefix string)
  (and (<= (length prefix)
           (length string))
       (loop for c1 across prefix
             for c2 across string
             always (char-equal c1 c2))))

(defmacro if (test then &body else)
  "Like CL:IF but with long ELSE branch"
  `(cl:if ,test ,then (locally ,@else)))

(eval-when (:compile-toplevel :load-toplevel)
  (cl:defun alist-get (key alist &optional default remove (testfn #'eq))
    "Find the first element of ALIST whose ‘car’ equals KEY and return its ‘cdr’.
If KEY is not found in ALIST, return DEFAULT.
Equality with KEY is tested by TESTFN, defaulting to ‘eq’.

You can use ‘alist-get’ in PLACE expressions.  This will modify
an existing association (more precisely, the first one if
multiple exist), or add a new element to the beginning of ALIST,
destructively modifying the list stored in ALIST.

When using it to set a value, optional argument REMOVE non-nil
means to remove KEY from ALIST if the new value is ‘eql’ to
DEFAULT (more precisely the first found association will be
deleted from the alist)."
    (declare (ignore remove))
    (let ((list (cl:assoc key alist :test testfn)))
      (declare (type list list))
      (if list (cdr list)
        default))))

(defmacro with-memoization (place &body forms)
  `(or ,place (setf ,place (locally ,@forms))))

(deftype constant-symbol () `(or (member t nil) keyword))

(deftype variable () `(and symbol (not constant-symbol)))

(eval-when (:compile-toplevel :load-toplevel)
  (defun uninterned-symbol-p (symbol)
    (declare (type symbol symbol))
    (let ((symbol-package (symbol-package symbol)))
      (and (not symbol-package)
           (loop for p in (delete symbol-package (list-all-packages) :test #'eq)
                 always (loop for s being the symbol in p
                              never (eq symbol s)))))))

(deftype uninterned-symbol () `(and symbol (satisfies uninterned-symbol-p)))

(defmacro fif (test f x)
  "If TEST evaluates to non-nil, return (F X).  Otherwise, return X."
  (let ((internal-function-name (make-symbol "FIF-VALUE")))
    `(flet ((,internal-function-name () ,x))
       (if ,test (funcall ,f (,internal-function-name))
         (,internal-function-name)))))

;; Support for common-lisp-indent-function in elisp
;; (put 'fif 'common-lisp-indent-function '(nil &body))

(defmacro wrap-if (condition wrapper var &optional (form nil form-supplied-p))
  (check-type var variable)
  `(fif ,condition (lambda (,var) ,wrapper) ,(if form-supplied-p form var)))

;; Support for common-lisp-indent-function in elisp
;; (put 'wrap-if 'common-lisp-indent-function '(nil nil &body))

(eval-when (:compile-toplevel :load-toplevel)
  (defun parse-binding (binding &optional default)
    (the (values variable t)
         (flet ((asserted-variable-symbol (symbol)
                  (declare (type symbol symbol))
                  (if (typep symbol `constant-symbol)
                      (error "Not a variable: ~S" symbol)
                    symbol))
                (asserted-variable-expr (expr)
                  (if (typep expr `variable) expr
                    (error "Not a variable: ~S" expr))))
           (declare (inline asserted-variable-symbol
                            asserted-variable-expr))
           (cond
             ((null binding)
              ;; (error "nil among bindings: ~S" bindings)
              (error "nil among bindings"))
             ((symbolp binding)
              (values (asserted-variable-symbol binding)
                      default))
             ((consp binding)
              (let ((parsing binding))
                (multiple-value-prog1 (values (asserted-variable-expr (pop parsing))
                                              (if (listp parsing) (pop parsing)
                                                (error "Illegal binding syntax for initial value in a binding ~S"
                                                       binding)))
                  (when parsing
                    (error "Illegal binding syntax; garbage after initial value in ~S"
                           binding)))))
             ((error "Illegal binding syntax; neither a variable, nor a cons: ~S"
                     binding)))))))

(defmacro with-parsed-binding ((var initial-value) binding &body body)
  (check-type var variable)
  (check-type initial-value variable)
  `(multiple-value-bind (,var ,initial-value) (parse-binding ,binding t)
     (declare (type variable ,var))
     ,@body))

(defmacro when-let* (bindings &body body &environment env)
  (declare (type list bindings))
  (if bindings
      (let (name nilp)
        (declare (type (or null uninterned-symbol) name)
                 (type (or null variable) nilp))
        (setq bindings
              (loop for b in bindings
                    collect (with-parsed-binding (var initial-value) b
                              `(,(the variable var)
                                ,(if (constantp initial-value env)
                                     (or (and (eval initial-value)
                                              initial-value)
                                         (progn (setq nilp var)
                                                (loop-finish)))
                                   `(or ,initial-value
                                        (return-from
                                         ,(with-memoization name
                                            (make-symbol "WHEN-LET*")))))))))
        (wrap-if name `(block ,name
                         ,let)
            let (flet ((note-unreachable ()
                         (format t "Deleting unreachable code in ~S as ~S is known to be nil~&"
                                 'when-let* (the variable nilp))))
                  (cond
                    (bindings
                     `(,(if (cdr bindings) 'let* 'let) ,bindings
                       ,@(if (null nilp) body
                           (note-unreachable)
                           `((declare (ignore ,@(mapcar #'car bindings)))))))
                    (nilp
                     (note-unreachable)
                     nil)
                    (:else (error "This should not happen"))))))
    `(locally ,@body)))

(define-setf-expander alist-get (key place
                                 &optional default remove (test #'eq)
                                 &environment env)
  (multiple-value-bind (syms vals newval setter getter) (get-setf-expansion
                                                         place env)
    (declare (ignore newval setter))
    (let ((counter 0)
          (remove-const-p (constantp remove env))
          (default-const-p (constantp default env)))
      (let ((o-o-key (prog1 (gensym "KEY-")
                       (setf counter (1- *gensym-counter*))))
            (o-o-place (let ((*gensym-counter* counter))
                         (gensym "PLACE-")))
            (o-o-default (unless default-const-p
                           (let ((*gensym-counter* counter))
                             (gensym "DEFAULT-"))))
            (o-o-remove (unless remove-const-p
                          (let ((*gensym-counter* counter))
                            (gensym "REMOVE-"))))
            (o-o-test (let ((*gensym-counter* counter))
                        (gensym "TEST-")))
            (value (let ((*gensym-counter* counter))
                     (gensym "VALUE-")))
            (next (let ((*gensym-counter* counter))
                    (gensym "NEXT-")))
            (tail (let ((*gensym-counter* counter))
                    (gensym "TAIL-")))
            (entry (let ((*gensym-counter* counter))
                     (gensym "ENTRY-"))))
        (values
         (append (list* o-o-key syms)
                 `( ,@(unless default-const-p `(,o-o-default))
                    ,@(unless remove-const-p `(,o-o-remove))
                    ,o-o-test))
         (append (list*     key vals)
                 `( ,@(unless default-const-p `(,default))
                    ,@(unless remove-const-p `(,remove))
                    ,test))
         (prog1 `(,value) (psetf key o-o-key place o-o-place
                                 test o-o-test)
                (psetf remove (if remove-const-p remove
                                o-o-remove)
                       default (if default-const-p default
                                 o-o-default)))
         (let ((removing-form
                 (unless (and remove-const-p (not remove))
                   `(let ((,place ,getter))
                      (declare (type list ,place))
                      (cond
                        ((null ,place))
                        ((funcall ,test ,key (caar ,place))
                         (pop ,getter))
                        (t
                         (when-let* ((,next (cdr ,place))
                                     (,entry
                                      (let ((,tail ,place))
                                        (declare (type list ,next))
                                        (declare (type cons ,tail))
                                        (loop :while ,next
                                              :do (locally (declare (type cons ,next))
                                                    (cl:if (funcall ,test ,key (caar ,next))
                                                           (return ,tail)
                                                           (pop ,tail)))
                                                  (setq ,next (cdr ,tail))))))
                           (pop (cdr ,entry)))))
                      ,value)))
               (adding-form `(let ((,entry (cl:assoc ,key ,getter :test ,test)))
                               (cl:if ,entry (rplacd ,entry ,value)
                                      (push (cons ,key ,value) ,getter))
                               ,value)))
           (cl:if remove-const-p
                  (cl:if remove
                         `(cl:if (eql ,value ,default)
                                 ,removing-form
                                 ,adding-form)
                         adding-form)
                  `(cl:if (and ,remove (eql ,value ,default))
                          ,removing-form
                          ,adding-form)))
         `(alist-get ,key ,getter ,default ,remove ,test))))))

(defmacro incf1 (x &optional (n 1))
  `(prog1 ,x (incf ,x ,n)))

(defmacro pnullf (&rest places)
  (when places `(psetf ,@(loop for x in places collect x collect nil))))

#+akater-types
(eval-when (:compile-toplevel :load-toplevel)
  (defvar type-checkers/list-of nil)
  (defvar type-checkers-names/list-of nil)
  (defvar type-checkers/list-of/counter 1)
  (defvar list-symbol-name "LIST")
  (defvar list-symbol (make-symbol "LIST"))

  (defun proper-list-with-elements-of-type-p (type list)
    ;; (declare (pure t) (side-effect-free t))
    (let ((result t))
      (loop while list
            unless (and (consp list)
                        (typep (pop list) type))
            do (pnullf list result))
      result)))

(eval-when (:compile-toplevel :load-toplevel)
  (defun partial (f &rest xs)
    (lambda (&rest post-xs) (apply f (append xs post-xs)))))

;; (deftype list-of (type)
;;   `(and list
;;         (satisfies
;;          ,(with-memoization (alist-get type type-checkers/list-of
;;                                        nil nil #'equal)
;;             (partial #'proper-list-with-elements-of-type-p type)))))

#+akater-types
(deftype list-of (type)
  `(and list
        (satisfies
         ,(with-memoization (alist-get type type-checkers-names/list-of
                                       nil nil #'equal)
            (let ((name ;; (make-symbol
                        ;;  (format nil "LIST-OF-TYPE-CHECKER-~S"
                        ;;          (incf1 type-checkers/list-of/counter)
                        ;;          ;; (incf1 *gensym-counter*)
                        ;;          ))
                        (intern
                         (format nil "LIST-OF-TYPE-CHECKER-~S"
                                 (incf1 type-checkers/list-of/counter)
                                 ;; (incf1 *gensym-counter*)
                                 )
                         (find-package "ORG-SYNTAX.SYSTEM"))
                        )
                  (lambda-expr
                   `(lambda (,list-symbol)
                      ,(format nil "Check ~A to be a (proper) list and check all its elements for type ~S"
                               (string-upcase list-symbol-name) type)
                      (proper-list-with-elements-of-type-p
                       ',type ,list-symbol))))
              (format t "Adding type checker ~S~&" name)
              (pushnew (cons name lambda-expr)
                       org-syntax.system::type-checkers/list-of
                       :key #'car :test #'eq)
              (compile name lambda-expr))))))

#+akater-types
(deftype alist (&optional key-type value-type)
  `(list-of (cons ,key-type ,value-type)))

(deftype index (&optional (start 0) (end '* end-supplied-p))
  `(and fixnum (integer ,start ,@(when end-supplied-p `(,end)))))

(defun string-prefix-whitespace-p (n string)
  "Whether there is enough whitespace at the beginning of STRING."
  (declare (type index n)
           (type string string))
  (loop for i below n
        for c across string
        always (whitespace-char-p c)))

(defun string-suffix-whitespace-p (start string)
  "Whether the rest of STRING is whitespace."
  (declare (type index start)
           (type string string))
  (loop for i from start
        for c across string
        always (whitespace-char-p c)))

(defun whitespace-char-p (char)
  (declare (type character char))
  (member char '(#\Space #\Tab) :test #'char-equal))

(defmacro awhen (test &body body)
  (if (null body) `(when ,test)
    `(let ((it ,test))
       (when it (locally ,@body)))))

(defmacro setq1 (place new)
  `(prog1 ,place (setq ,place ,new)))

(defmacro o-o (var &body format-args)
  "Return expression representing binding a fresh gensym to VAR; ressaign VAR to it.

Supply FORMAT-ARGS to specify symbol name (at runtime).

When FORMAT-ARGS is (t), construct name as `gensym' would.

This macro is used to once-onlyize variables in macros."
  ;; (check-type var variable)
  `(let ((gensym (make-symbol
                  ,(if format-args
                       `(format nil ,@(if (and (null (cdr format-args))
                                               (eq t (car format-args)))
                                          `("~S-~S" ',var (incf1 gensym-counter))
                                        format-args))
                     (symbol-name var)))))
     `(,gensym ,(setq1 ,var gensym))))

(defmacro aif (test then &body else)
  (if (null else) `(awhen ,test ,then)
      `(let (,(o-o test))
         (if ,test (let ((it ,test))
                     (declare (type (not null) it))
                     ,then)
             (locally ,@else)))))

(defmacro when1 (test &body body)
  `(let (,(o-o test))
     (when ,test
       (locally (declare (type (not null) ,test))
         ,@body ,test))))

(defmacro nullf (&rest places)
  (when places `(setf ,@(loop for x in places collect x collect nil))))

(defmacro zerof (&rest places)
  (when places `(setf ,@(loop for x in places collect x collect 0))))

(eval-when (:compile-toplevel :load-toplevel)
  (defun make-aand-form (xs)
    (declare (type cons xs))
    `(and it
          ,@(loop while (cdr xs)
                  collect `(setq it ,(pop xs)))
          ,(car xs))))

(defmacro aand (&rest xs)
  (cond
    ((null xs))
    ((null (cdr xs)) (car xs))
    (`(let ((it ,(pop xs)))
        ,(make-aand-form xs)))))

(defmacro aand1 (x &rest xs)
  (if (null xs) x
      (let ((not-resettable (make-symbol "X")))
        `(let* ((,not-resettable ,x)
                (it ,not-resettable))
           (when ,(make-aand-form xs)
             ,not-resettable)))))

(eval-when (:compile-toplevel :load-toplevel)
  (defun variablep (x)
    (when (typep x `variable)
      x)))

(defmacro when-let (bindings &body body &environment env)
  (declare (type list bindings))
  (if bindings
      (let (name nilp)
        (declare (type (or null uninterned-symbol) name)
                 (type (or null variable) nilp))
        (setq bindings
              (loop for b in bindings
                    collect (with-parsed-binding (var initial-value) b
                              `(,(the variable var)
                                ,(if (constantp initial-value env)
                                     (or (and (eval initial-value)
                                              initial-value)
                                         (progn (setq nilp var)
                                                (loop-finish)))
                                   `(or ,initial-value
                                        (return-from
                                         ,(with-memoization name
                                            (make-symbol "WHEN-LET*")))))))))
        (wrap-if name `(block ,name
                         ,let)
            let (flet ((note-unreachable ()
                         (format t "Deleting unreachable code in ~S as ~S is known to be nil~&"
                          'when-let (the variable nilp))))
                  (cond
                    (bindings
                     `(let ,bindings
                        ,@(if (null nilp) body
                            (note-unreachable)
                            `((declare (ignore ,@(mapcar #'car bindings)))))))
                    (nilp
                     (note-unreachable)
                     nil)
                    (:else (error "This should not happen"))))))
    `(locally ,@body)))

(deftype string-designator () `(or symbol string character))

(define-modify-macro nreversef () nreverse)

(defmacro nullf1 (x) `(prog1 ,x (nullf ,x)))

(defun position-not (item seq &key (test #'eql) (start 0))
  (declare (type (or list vector) seq))
  (position item seq :test-not test :start start))

(defmacro plist-with-keys (&rest vars)
  `(list ,@(loop for v in vars collect (intern (symbol-name v) "KEYWORD")
                 collect v)))

(define-modify-macro carf () car)

(defun ensure-string (string-designator)
  (declare (type string-designator string-designator))
  (funcall (etypecase string-designator
             (symbol #'symbol-name)
             (string #'identity)
             (character #'string))
           string-designator))

(defmacro until (test &body body)
  `(loop while (not ,test) do ,@body))

(defun nest (n f x)
  (check-type n (integer 0))
  (check-type f function)
  (until (zerop n)
    (psetf x (funcall f x)
           n (1- n)))
  x)

